/*
 *  GOL - Game of life simulation software
 *  Copyright (C) 1996 Michael Raymond Brown
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Author: Michael Raymond Brown
 *  EMail:  michael@arrakeen.demon.co.uk
 *  WWW:    http://www.arrakeen.demon.co.uk/
 *  Address:  Michael Brown
 *            Kia Ora
 *            Brook Street
 *            Great Bromley
 *            Nr Colchester
 *            Essex
 *            CO7 7HX
 *            England
 */


#ifdef __GNUC__
#pragma implementation
#endif

#include "life.h"
#include <string.h>
#include <iostream.h>


Life::Life()
   {
   }
Life::~Life()
   {
   }

#define HSTEP	4
#define VSTEP	(BITSPERINT/4)


	// I'm ignoring `wrap' for now
local_Life::local_Life(int w, int h, int pwrap
		     , char prules[10], char pmask[3][3])
   {int s;

    rules[ 0]=pwrap?'y':'n';
    rules[ 1]='-';
    rules[ 2]=pmask[0][0]?'x':' ';
    rules[ 3]=pmask[1][0]?'x':' ';
    rules[ 4]=pmask[2][0]?'x':' ';
    rules[ 5]=pmask[0][1]?'x':' ';
    rules[ 6]=pmask[1][1]?'x':' ';
    rules[ 7]=pmask[2][1]?'x':' ';
    rules[ 8]=pmask[0][2]?'x':' ';
    rules[ 9]=pmask[1][2]?'x':' ';
    rules[10]=pmask[2][2]?'x':' ';
    rules[11]='-';
    rules[12]=(prules[0]==CLEAR)?'c':((prules[0]==LEAVE)?'l':'s');
    rules[13]=(prules[1]==CLEAR)?'c':((prules[1]==LEAVE)?'l':'s');
    rules[14]=(prules[2]==CLEAR)?'c':((prules[2]==LEAVE)?'l':'s');
    rules[15]=(prules[3]==CLEAR)?'c':((prules[3]==LEAVE)?'l':'s');
    rules[16]=(prules[4]==CLEAR)?'c':((prules[4]==LEAVE)?'l':'s');
    rules[17]=(prules[5]==CLEAR)?'c':((prules[5]==LEAVE)?'l':'s');
    rules[18]=(prules[6]==CLEAR)?'c':((prules[6]==LEAVE)?'l':'s');
    rules[19]=(prules[7]==CLEAR)?'c':((prules[7]==LEAVE)?'l':'s');
    rules[20]=(prules[8]==CLEAR)?'c':((prules[8]==LEAVE)?'l':'s');
    rules[21]=(prules[9]==CLEAR)?'c':((prules[9]==LEAVE)?'l':'s');

     width = (w+63)&~63;
    height = (h+63)&~63;

     scan_width = (width/HSTEP) + 1;
    scan_height = (height/VSTEP);

    s = scan_width * (scan_height + 2);
    field = new u_int[ s ];
    memset(field, 0, s*sizeof(int));

    spare = field + (scan_width * (scan_height+1));
    reverse=0; // This means move the display down while ticking
	       // After a tick this value is inverted, and field
	       // is modified to ensure plot/point still work
    if (pwrap)
	 tickf=&tick_upw;
    else tickf=&tick_up;

    linespertick = (128*1024)/(width*(BITSPERINT/HSTEP));
    if (!linespertick) linespertick=1;
    scan_count=-1;
    wrap=pwrap;

   {u_char temp[2048];
    int   count;
    int  mask=0;

    if (pmask[0][0]) mask =0x001;
    if (pmask[1][0]) mask|=0x002;
    if (pmask[2][0]) mask|=0x004;
    if (pmask[0][1]) mask|=0x010;
    if (pmask[1][1]) mask|=0x020;
    if (pmask[2][1]) mask|=0x040;
    if (pmask[0][2]) mask|=0x100;
    if (pmask[1][2]) mask|=0x200;
    if (pmask[2][2]) mask|=0x400;

    count=2047; do {
	s = count & mask;
	s = ((s >> 1) & 0x5555) + (s & 0x5555);
	s = ((s >> 2) & 0x3333) + (s & 0x3333);
	s = ((s >> 4) & 0x0f0f) + (s & 0x0f0f);
	s = ((s >> 8) & 0x00ff) + (s & 0x00ff);
	switch (prules[s])
	   {case CLEAR:
		temp[count]=0;
		break;
	    case LEAVE:
		temp[count]=(count>>5)&1;
		break;
	    case SET:
		temp[count]=1;
		break;
	   }
    } while ((--count) >= 0);

    count = 65535; do {
	lookup[count]=(temp[(count>>0)&0x777]<<0)
		    |(temp[(count>>1)&0x777]<<1)
		    |(temp[(count>>4)&0x777]<<4)
		    |(temp[(count>>5)&0x777]<<5);
    } while ((--count) >= 0);

   }}

local_Life::~local_Life()
   {if (field) delete (field - (reverse?scan_width:0));
   }


inline void local_Life::tick_scan(u_int v1, u_int v2, u_int v3, const u_int* p1, const u_int* p2, const u_int* p3, u_int* dst)
   {u_int w1, w2, rslt;
    u_int tmp1, tmp2;
    u_int i = 0;
    int count = width/HSTEP;
u_int mask_11 = 0x11111111;
u_int mask_ee = 0xeeeeeeee;
u_int mask_77 = 0x77777777;
u_int mask_88 = 0x88888888;
u_int mask_16 = 0xffff;

    dst--;
    v1 = (v1&0xf0000000)|(v3&0xf);
    w1 = (p1[i]&0xf0000000)|(p3[i]&0xf);
    w2 = p2[i];
    i++;

    do {int s;
#if BITSPERINT == 32
	// Do left
	tmp1 = ((v1 >> 3) & mask_11) | ((w1 << 1) & mask_ee);
	tmp2 = ((v2 >> 3) & mask_11) | ((w2 << 1) & mask_ee);
	rslt  = (u_int(lookup[((tmp1>>28)|(tmp2<<4))&mask_16]))
	      | (u_int(lookup[(tmp2>> 4)&mask_16])<< 8)
	      | (u_int(lookup[(tmp2>>12)&mask_16])<<16)
	      | (u_int(lookup[((tmp1&0xf)<<12)|(tmp2>>20)])<<24);

	v1=w1; v2=w2;
	w1 = (p1[i]&0xf0000000)|(p3[i]&0xf);
	w2=p2[i];

	// Do right
	tmp1 = ((v1 >> 1) & mask_77) | ((w1 << 3) & mask_88);
	tmp2 = ((v2 >> 1) & mask_77) | ((w2 << 3) & mask_88);
	rslt |= (u_int(lookup[((tmp1>>28)|(tmp2<<4))&mask_16])<<2)
	      | (u_int(lookup[(tmp2>> 4)&mask_16])<<10)
	      | (u_int(lookup[(tmp2>>12)&mask_16])<<18)
	      | (u_int(lookup[((tmp1&0xf)<<12)|(tmp2>>20)])<<26);

	dst[i]=rslt;
	i++;
#else
#error Not implemented
#endif

    } while (--count);
   }



  /** Scans go down, the target is 1 above the current scan line **/
int local_Life::tick_down()
   {int ticks = linespertick+1;
	// +1 adjusts for the fact that i pre-decrement the tick counter

    if (scan_count == -1)
       {
	p1 = spare;
	p2 = field;
	p3 = field+scan_width;
	dst = field-scan_width;
	scan_count = scan_height-1;
       }

    do {
	if (!--ticks) return 0;
	tick_scan(0,0,0,p1,p2,p3,dst);
	p1=p2; p2=p3;
	p3+=scan_width;
	dst+=scan_width;
    } while (--scan_count);
    tick_scan(0,0,0,p1,p2,spare,dst);

    scan_count=-1;
    field-=scan_width;
    reverse=~reverse;
    tickf=&tick_up;
    return 1;
   }
  /** Scans go up, the target is 1 below the current scan line **/
int local_Life::tick_up()
   {int ticks = linespertick+1;
	// +1 adjusts for the fact that i pre-decrement the tick counter

    if (scan_count == -1)
       {
	p1 = field+(scan_width*(scan_height-2));
	p2 = p1+scan_width;
	p3 = spare;
	dst = p2+scan_width;
	scan_count = scan_height-1;
       }

    do {
	if (!--ticks) return 0;
	tick_scan(0,0,0,p1,p2,p3,dst);
	p3=p2; p2=p1;
	p1-=scan_width;
	dst-=scan_width;
    } while (--scan_count);
    tick_scan(0,0,0,spare,p2,p3,dst);

    scan_count=-1;
    field+=scan_width;
    reverse=~reverse;
    tickf=&tick_down;
    return 1;
   }
int local_Life::tick_downw()
   {int scanw = scan_width-2;
    int ticks = linespertick+1;
	// +1 adjusts for the fact that i pre-decrement the tick counter

    if (scan_count == -1)
       {int count;
	p1 = field+(scan_width*(scan_height-1));
	p2 = field;
	p3 = field+scan_width;
	scan_count = scan_height-1;

	dst = field;
	count=scan_height; do {
	    u_int v=dst[0];
	    dst+=scan_width;
	    dst[-1]=v;
	} while (--count);
	dst = p2-scan_width;
	memcpy(spare, p2, scan_width*sizeof(u_int));
       }

    do {
	if (!--ticks) return 0;
	tick_scan(p1[scanw],p2[scanw],p3[scanw],p1,p2,p3,dst);
	p1=p2; p2=p3;
	p3+=scan_width;
	dst+=scan_width;
    } while (--scan_count);
    tick_scan(p1[scanw],p2[scanw],spare[scanw],p1,p2,spare,dst);

    scan_count=-1;
    field-=scan_width;
    reverse=~reverse;
    tickf=&tick_upw;
    return 1;
   }
int local_Life::tick_upw()
   {int scanw = scan_width-2;
    int ticks = linespertick+1;
	// +1 adjusts for the fact that i pre-decrement the tick counter

    if (scan_count == -1)
       {int count;
	p1 = field+(scan_width*(scan_height-2));
	p2 = p1+scan_width;
	p3 = field;
	scan_count = scan_height-1;

	dst = field;
	count=scan_height; do {
	    u_int v=dst[0];
	    dst+=scan_width;
	    dst[-1]=v;
	} while (--count);
	dst = p2+scan_width;
	memcpy(spare, p2, scan_width*sizeof(u_int));
       }

    do {
	if (!--ticks) return 0;
	tick_scan(p1[scanw],p2[scanw],p3[scanw],p1,p2,p3,dst);
	p3=p2; p2=p1;
	p1-=scan_width;
	dst-=scan_width;
    } while (--scan_count);
    tick_scan(spare[scanw],p2[scanw],p3[scanw],spare,p2,p3,dst);

    scan_count=-1;
    field+=scan_width;
    reverse=~reverse;
    tickf=&tick_downw;
    return 1;
   }



int local_Life::tick()
   {return (this->*tickf)();
   }

int local_Life::isOk()
   {return scan_count==-1;
   }

void local_Life::plot(int x, int y, int value)
   {if (scan_count!=-1 || x<0 || y<0 || x>=width || y>=height) return;
    if (value)
         field[scan_width*(y/VSTEP)+(x/HSTEP)] |= 1<<(4*(y%VSTEP)+(x%HSTEP));
    else field[scan_width*(y/VSTEP)+(x/HSTEP)] &= ~(1<<(4*(y%VSTEP)+(x%HSTEP)));
   }
int local_Life::point(int x, int y)
   {if (scan_count!=-1 || x<0 || y<0 || x>=width || y>=height) return 0;
    return (field[scan_width*(y/VSTEP)+(x/HSTEP)] >> (4*(y%VSTEP)+(x%HSTEP))) & 1;
   }
int local_Life::getWidth()
   {return width;
   }
int local_Life::getHeight()
   {return height;
   }


void local_Life::getArea(int x, int y, u_int* dst)
   {
    u_int v1, v2;
    u_int v3, v4;
    u_int*    p1;
    u_int   mask;

    if (scan_count != -1) return;
    if (x<0 || y<0 || (x*16)>=width || (y*16)>=height)
       {
	dst[ 0]=0; dst[ 1]=0; dst[ 2]=0; dst[ 3]=0;
	dst[ 4]=0; dst[ 5]=0; dst[ 6]=0; dst[ 7]=0;
	dst[ 8]=0; dst[ 9]=0; dst[10]=0; dst[11]=0;
	dst[12]=0; dst[13]=0; dst[14]=0; dst[15]=0;
	return;
       }

#if BITSPERINT == 32
    p1 = field + (x*4) + (scan_width * y * 2);
#else
    p1 = field + (x*4) + (scan_width * y);
#endif


    mask = 0x0000000f;
    v1=p1[0]; v2=p1[1]; v3=p1[2]; v4=p1[3];
    dst[ 0] = ((v1&mask)    ) | ((v2&mask)<< 4) | ((v3&mask)<< 8) | ((v4&mask)<<12); mask<<=4;
    dst[ 1] = ((v1&mask)>> 4) | ((v2&mask)    ) | ((v3&mask)<< 4) | ((v4&mask)<< 8); mask<<=4;
    dst[ 2] = ((v1&mask)>> 8) | ((v2&mask)>> 4) | ((v3&mask)    ) | ((v4&mask)<< 4); mask<<=4;
    dst[ 3] = ((v1&mask)>>12) | ((v2&mask)>> 8) | ((v3&mask)>> 4) | ((v4&mask)    ); mask<<=4;
    dst[ 4] = ((v1&mask)>>16) | ((v2&mask)>>12) | ((v3&mask)>> 8) | ((v4&mask)>> 4); mask<<=4;
    dst[ 5] = ((v1&mask)>>20) | ((v2&mask)>>16) | ((v3&mask)>>12) | ((v4&mask)>> 8); mask<<=4;
    dst[ 6] = ((v1&mask)>>24) | ((v2&mask)>>20) | ((v3&mask)>>16) | ((v4&mask)>>12); mask<<=4;
    dst[ 7] = ((v1&mask)>>28) | ((v2&mask)>>24) | ((v3&mask)>>20) | ((v4&mask)>>16);
#if BITSPERINT == 32
    mask = 0x0000000f; p1+=scan_width;
    v1=p1[0]; v2=p1[1]; v3=p1[2]; v4=p1[3];
    dst[ 8] = ((v1&mask)    ) | ((v2&mask)<< 4) | ((v3&mask)<< 8) | ((v4&mask)<<12); mask<<=4;
    dst[ 9] = ((v1&mask)>> 4) | ((v2&mask)    ) | ((v3&mask)<< 4) | ((v4&mask)<< 8); mask<<=4;
    dst[10] = ((v1&mask)>> 8) | ((v2&mask)>> 4) | ((v3&mask)    ) | ((v4&mask)<< 4); mask<<=4;
    dst[11] = ((v1&mask)>>12) | ((v2&mask)>> 8) | ((v3&mask)>> 4) | ((v4&mask)    ); mask<<=4;
    dst[12] = ((v1&mask)>>16) | ((v2&mask)>>12) | ((v3&mask)>> 8) | ((v4&mask)>> 4); mask<<=4;
    dst[13] = ((v1&mask)>>20) | ((v2&mask)>>16) | ((v3&mask)>>12) | ((v4&mask)>> 8); mask<<=4;
    dst[14] = ((v1&mask)>>24) | ((v2&mask)>>20) | ((v3&mask)>>16) | ((v4&mask)>>12); mask<<=4;
    dst[15] = ((v1&mask)>>28) | ((v2&mask)>>24) | ((v3&mask)>>20) | ((v4&mask)>>16);
#else
    mask <<= 4;
    dst[ 8] = ((v1&mask)>>32) | ((v2&mask)>>28) | ((v3&mask)>>24) | ((v4&mask)>>20); mask<<=4;
    dst[ 9] = ((v1&mask)>>36) | ((v2&mask)>>32) | ((v3&mask)>>28) | ((v4&mask)>>24); mask<<=4;
    dst[10] = ((v1&mask)>>40) | ((v2&mask)>>36) | ((v3&mask)>>32) | ((v4&mask)>>28); mask<<=4;
    dst[11] = ((v1&mask)>>44) | ((v2&mask)>>40) | ((v3&mask)>>36) | ((v4&mask)>>32); mask<<=4;
    dst[12] = ((v1&mask)>>48) | ((v2&mask)>>44) | ((v3&mask)>>40) | ((v4&mask)>>36); mask<<=4;
    dst[13] = ((v1&mask)>>52) | ((v2&mask)>>48) | ((v3&mask)>>44) | ((v4&mask)>>40); mask<<=4;
    dst[14] = ((v1&mask)>>56) | ((v2&mask)>>52) | ((v3&mask)>>48) | ((v4&mask)>>44); mask<<=4;
    dst[15] = ((v1&mask)>>60) | ((v2&mask)>>56) | ((v3&mask)>>52) | ((v4&mask)>>48);
#endif
   }



void local_Life::setArea(int x, int y, const u_int* src)
   {u_int* dst;
    u_int mask;

    if (scan_count!=-1 || x<0 || y<0 || (x*16)>=width || (y*16)>=height)
	return;

#if BITSPERINT == 32
    dst = field + (x*4) + (scan_width * y * 2);
#else
    dst = field + (x*4) + (scan_width * y);
#endif

#if BITSPERINT == 32
    mask=15;
    dst[0]=((src[ 0]&mask)    )|((src[ 1]&mask)<< 4)|((src[ 2]&mask)<< 8)|((src[ 3]&mask)<<12)
	  |((src[ 4]&mask)<<16)|((src[ 5]&mask)<<20)|((src[ 6]&mask)<<24)|((src[ 7]&mask)<<28);
    mask<<=4;
    dst[1]=((src[ 0]&mask)>> 4)|((src[ 1]&mask)    )|((src[ 2]&mask)<< 4)|((src[ 3]&mask)<< 8)
	  |((src[ 4]&mask)<<12)|((src[ 5]&mask)<<16)|((src[ 6]&mask)<<20)|((src[ 7]&mask)<<24);
    mask<<=4;
    dst[2]=((src[ 0]&mask)>> 8)|((src[ 1]&mask)>> 4)|((src[ 2]&mask)    )|((src[ 3]&mask)<< 4)
	  |((src[ 4]&mask)<< 8)|((src[ 5]&mask)<<12)|((src[ 6]&mask)<<16)|((src[ 7]&mask)<<20);
    mask<<=4;
    dst[3]=((src[ 0]&mask)>>12)|((src[ 1]&mask)>> 8)|((src[ 2]&mask)>> 4)|((src[ 3]&mask)    )
	  |((src[ 4]&mask)<< 4)|((src[ 5]&mask)<< 8)|((src[ 6]&mask)<<12)|((src[ 7]&mask)<<16);
    dst+=scan_width;
    mask=15;
    dst[0]=((src[ 8]&mask)    )|((src[ 9]&mask)<< 4)|((src[10]&mask)<< 8)|((src[11]&mask)<<12)
	  |((src[12]&mask)<<16)|((src[13]&mask)<<20)|((src[14]&mask)<<24)|((src[15]&mask)<<28);
    mask<<=4;
    dst[1]=((src[ 8]&mask)>> 4)|((src[ 9]&mask)    )|((src[10]&mask)<< 4)|((src[11]&mask)<< 8)
	  |((src[12]&mask)<<12)|((src[13]&mask)<<16)|((src[14]&mask)<<20)|((src[15]&mask)<<24);
    mask<<=4;
    dst[2]=((src[ 8]&mask)>> 8)|((src[ 9]&mask)>> 4)|((src[10]&mask)    )|((src[11]&mask)<< 4)
	  |((src[12]&mask)<< 8)|((src[13]&mask)<<12)|((src[14]&mask)<<16)|((src[15]&mask)<<20);
    mask<<=4;
    dst[3]=((src[ 8]&mask)>>12)|((src[ 9]&mask)>> 8)|((src[10]&mask)>> 4)|((src[11]&mask)    )
	  |((src[12]&mask)<< 4)|((src[13]&mask)<< 8)|((src[14]&mask)<<12)|((src[15]&mask)<<16);

#else
    mask=15;
    dst[0]=((src[ 0]&mask)    )|((src[ 1]&mask)<< 4)|((src[ 2]&mask)<< 8)|((src[ 3]&mask)<<12)
	  |((src[ 4]&mask)<<16)|((src[ 5]&mask)<<20)|((src[ 6]&mask)<<24)|((src[ 7]&mask)<<28)
	  |((src[ 8]&mask)<<32)|((src[ 9]&mask)<<36)|((src[10]&mask)<<40)|((src[11]&mask)<<44)
	  |((src[12]&mask)<<48)|((src[13]&mask)<<52)|((src[14]&mask)<<56)|((src[15]&mask)<<60);
    mask<<=4;
    dst[1]=((src[ 0]&mask)>> 4)|((src[ 1]&mask)    )|((src[ 2]&mask)<< 4)|((src[ 3]&mask)<< 8)
	  |((src[ 4]&mask)<<12)|((src[ 5]&mask)<<16)|((src[ 6]&mask)<<20)|((src[ 7]&mask)<<24)
	  |((src[ 8]&mask)>>28)|((src[ 9]&mask)<<32)|((src[10]&mask)<<36)|((src[11]&mask)<<40)
	  |((src[12]&mask)<<44)|((src[13]&mask)<<48)|((src[14]&mask)<<52)|((src[15]&mask)<<56);
    mask<<=4;
    dst[2]=((src[ 0]&mask)>> 8)|((src[ 1]&mask)>> 4)|((src[ 2]&mask)    )|((src[ 3]&mask)<< 4)
	  |((src[ 4]&mask)<< 8)|((src[ 5]&mask)<<12)|((src[ 6]&mask)<<16)|((src[ 7]&mask)<<20)
	  |((src[ 8]&mask)>>24)|((src[ 9]&mask)<<28)|((src[10]&mask)<<32)|((src[11]&mask)<<36)
	  |((src[12]&mask)<<40)|((src[13]&mask)<<44)|((src[14]&mask)<<48)|((src[15]&mask)<<50);
    mask<<=4;
    dst[3]=((src[ 0]&mask)>>12)|((src[ 1]&mask)>> 8)|((src[ 2]&mask)>> 4)|((src[ 3]&mask)    )
	  |((src[ 4]&mask)<< 4)|((src[ 5]&mask)<< 8)|((src[ 6]&mask)<<12)|((src[ 7]&mask)<<16)
	  |((src[ 8]&mask)>>20)|((src[ 9]&mask)<<24)|((src[10]&mask)<<28)|((src[11]&mask)<<32)
	  |((src[12]&mask)<<36)|((src[13]&mask)<<40)|((src[14]&mask)<<44)|((src[15]&mask)<<48);

#endif

   }



#define irand(val, count)\
        (val+=(val>>1)+(++count))

void local_Life::randomize()
   {u_int* ptr=field;
    int   x, y;
    static u_int rcount = 89234;
    static u_int rvalue = 8324444;

    if (scan_count!=-1) return;

    for (y=scan_height; y--; )
	{for (x=scan_width-1; x--; )
	      *ptr++=irand(rvalue,rcount);
	 ptr++;
	}
   }

void local_Life::clear()
   {u_int* ptr=field;
    int  x, y;

    if (scan_count!=-1) return;

    for (y=scan_height; y--; )
	{for (x=scan_width-1; x--; ) *ptr++=0;
	 ptr++;
	}
   }

